use crate::rstd::{cast::*, typeless::*};

use super::*;

struct TracedResolver<'a, Ctx: 'a + Context> {
    resolver: Rc<dyn Resolver<'a, Ctx>>,
}

impl<'a, Ctx: 'a + Context<T = TracedClass>> TracedResolver<'a, Ctx> {
    fn wrap(resolver: Rc<dyn Resolver<'a, Ctx>>) -> Rc<dyn Resolver<'a, Ctx>> {
        Rc::new(Self { resolver })
    }
}

impl<'a, Ctx: 'a + Context<T = TracedClass>> Resolver<'a, Ctx> for TracedResolver<'a, Ctx> {
    fn resolve(self: Rc<Self>, address: Address) -> HashResolution<'a, Ctx> {
        TracedClass::fmap(
            |resolved| {
                let (src, resolver) = resolved?;
                let delayed: Rc<dyn Resolver<'a, Ctx>> = Rc::new(TracedResolver { resolver });
                Ok((src, delayed))
            },
            self.resolver.clone().resolve(address),
        )
        .after_resolution()
    }
}

/// Extension trait to trace the evaluation flow.
pub trait Traceable<'a, Ctx: 'a + Context<T = TracedClass>>: Mentionable<'a, Ctx> + Sized {
    /// Re-cast the value, adding an extra[^extra]
    /// note ([`RenderedCommon::Resolution`]) to the trace on each resolution.
    ///
    /// [^extra]: applying [`Traceable::trace`] multiple times
    /// might affect the trace in undesireable ways
    fn trace(self: Rc<Self>) -> CastResult<'a, Ctx, Self>;
}

impl<'a, Ctx: 'a + Context<T = TracedClass>, A: Mentionable<'a, Ctx>> Traceable<'a, Ctx> for A
where
    Ctx::LookupError<'a>: From<CastError<'a>>,
{
    fn trace(self: Rc<Self>) -> CastResult<'a, Ctx, Self> {
        let factory = self.factory();
        TypelessMentionable::from_typed(self).cast_full(factory, TracedResolver::wrap)
    }
}
